r"""
Dynamical semigroups

A dynamical semigroup is a finitely generated subsemigroup of
the endomorphism ring of a subscheme of projective or affine space.

AUTHORS:

 - Dang Phan (August 6th, 2023): initial implementation
"""

# ****************************************************************************
# Dang Phan <dang8phan@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from collections.abc import Collection
from sage.categories.fields import Fields
from sage.categories.number_fields import NumberFields
from sage.categories.semigroups import Semigroups
from sage.dynamics.arithmetic_dynamics.affine_ds import DynamicalSystem_affine
from sage.dynamics.arithmetic_dynamics.generic_ds import DynamicalSystem
from sage.dynamics.arithmetic_dynamics.projective_ds import DynamicalSystem_projective
from sage.misc.classcall_metaclass import typecall
from sage.misc.inherit_comparison import InheritComparisonClasscallMetaclass
from sage.rings.finite_rings.finite_field_base import FiniteField
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.parent import Parent


class DynamicalSemigroup(Parent, metaclass=InheritComparisonClasscallMetaclass):
    r"""
    A dynamical semigroup defined by a multiple dynamical systems on projective or affine space.

    INPUT:

    - ``ds_data`` -- list or tuple of dynamical systems or objects that define dynamical systems

    OUTPUT:

    :class:`DynamicalSemigroup_affine` if ``ds_data`` only contains dynamical systems
    over affine space; and :class:`DynamicalSemigroup_projective` otherwise.

    EXAMPLES::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: DynamicalSemigroup(([x, y], [x^2, y^2]))
        Dynamical semigroup over Projective Space of dimension 1 over Rational Field
         defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    ::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([x^2, y^2], P)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over Rational Field
         defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    ::

        sage: A.<x> = AffineSpace(QQ, 1)
        sage: f = DynamicalSystem_affine(x, A)
        sage: DynamicalSemigroup(f)
        Dynamical semigroup over Affine Space of dimension 1 over Rational Field
         defined by 1 dynamical system:
          Dynamical System of Affine Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x) to (x)

    ::

        sage: A.<x> = AffineSpace(QQ, 1)
        sage: f = DynamicalSystem(x, A)
        sage: g = DynamicalSystem(x^2, A)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Affine Space of dimension 1 over Rational Field
         defined by 2 dynamical systems:
          Dynamical System of Affine Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x) to (x)
          Dynamical System of Affine Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x) to (x^2)

    ::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: X = P.subscheme(x - y)
        sage: f = DynamicalSystem_projective([x, y], X)
        sage: g = DynamicalSystem_projective([x^2, y^2], X)
        sage: DynamicalSemigroup_projective([f, g])
        Dynamical semigroup over Closed subscheme of Projective Space of dimension 1
         over Rational Field defined by: x - y
         defined by 2 dynamical systems:
          Dynamical System of Closed subscheme of Projective Space of dimension 1
           over Rational Field defined by: x - y
             Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Closed subscheme of Projective Space of dimension 1
           over Rational Field defined by: x - y
             Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    If a dynamical semigroup is built from dynamical systems with different base rings,
    all systems will be coerced to the largest base ring::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: Q.<z,w> = ProjectiveSpace(RR, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([z^2, w^2], Q)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over
         Real Field with 53 bits of precision defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1
           over Real Field with 53 bits of precision
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1
           over Real Field with 53 bits of precision
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    ::

        sage: A.<x> = AffineSpace(QQ, 1)
        sage: B.<y> = AffineSpace(RR, 1)
        sage: f = DynamicalSystem(x, A)
        sage: g = DynamicalSystem(y^2, B)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Affine Space of dimension 1 over
         Real Field with 53 bits of precision defined by 2 dynamical systems:
          Dynamical System of Affine Space of dimension 1 over
           Real Field with 53 bits of precision
            Defn: Defined on coordinates by sending (x) to (x)
          Dynamical System of Affine Space of dimension 1 over
           Real Field with 53 bits of precision
            Defn: Defined on coordinates by sending (x) to (x^2)

    If a dynamical semigroup is built from dynamical systems over number fields, a composite number field is created
    and all systems will be coerced to it. This composite number field contains all of the initial number fields::

        sage: # needs sage.rings.number_field
        sage: R.<r> = QQ[]
        sage: K.<k> = NumberField(r^2 - 2)
        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: Q.<x,y> = ProjectiveSpace(K, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([z^2, w^2], Q)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over
         Number Field in k with defining polynomial r^2 - 2 defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Number Field in k with defining polynomial r^2 - 2
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over
           Number Field in k with defining polynomial r^2 - 2
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    ::

        sage: # needs sage.rings.number_field
        sage: R.<r> = QQ[]
        sage: K.<k> = NumberField(r^2 - 2)
        sage: L.<l> = NumberField(r^2 - 3)
        sage: P.<x,y> = ProjectiveSpace(K, 1)
        sage: Q.<z,w> = ProjectiveSpace(L, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([z^2, w^2], Q)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over
         Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
         defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over
           Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    ::

        sage: # needs sage.rings.number_field
        sage: R.<r> = QQ[]
        sage: K.<k> = NumberField(r^2 - 2)
        sage: L.<l> = NumberField(r^2 - 3)
        sage: P.<x> = AffineSpace(K, 1)
        sage: Q.<y> = AffineSpace(L, 1)
        sage: f = DynamicalSystem(x, P)
        sage: g = DynamicalSystem(y^2, Q)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Affine Space of dimension 1 over
         Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
         defined by 2 dynamical systems:
          Dynamical System of Affine Space of dimension 1 over
           Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
            Defn: Defined on coordinates by sending (x) to (x)
          Dynamical System of Affine Space of dimension 1 over
           Number Field in kl with defining polynomial r^4 - 10*r^2 + 1
            Defn: Defined on coordinates by sending (x) to (x^2)

    A dynamical semigroup may contain dynamical systems over function fields::

        sage: R.<r> = QQ[]
        sage: P.<x,y> = ProjectiveSpace(R, 1)
        sage: f = DynamicalSystem([r * x, y], P)
        sage: g = DynamicalSystem([x, r * y], P)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over Univariate
         Polynomial Ring in r over Rational Field defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Univariate Polynomial Ring in r over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (r*x : y)
          Dynamical System of Projective Space of dimension 1 over
           Univariate Polynomial Ring in r over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : r*y)

    ::

        sage: R.<r> = QQ[]
        sage: P.<x,y> = ProjectiveSpace(R, 1)
        sage: f = DynamicalSystem([r * x, y], P)
        sage: g = DynamicalSystem([x, y], P)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over Univariate
         Polynomial Ring in r over Rational Field defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Univariate Polynomial Ring in r over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (r*x : y)
          Dynamical System of Projective Space of dimension 1 over
           Univariate Polynomial Ring in r over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : y)

    ::

        sage: R.<r,s> = QQ[]
        sage: P.<x,y> = ProjectiveSpace(R, 1)
        sage: f = DynamicalSystem([r * x, y], P)
        sage: g = DynamicalSystem([s * x, y], P)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over Multivariate
         Polynomial Ring in r, s over Rational Field defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Multivariate Polynomial Ring in r, s over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (r*x : y)
          Dynamical System of Projective Space of dimension 1 over
           Multivariate Polynomial Ring in r, s over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (s*x : y)

    ::

        sage: R.<r,s> = QQ[]
        sage: P.<x,y> = ProjectiveSpace(R, 1)
        sage: f = DynamicalSystem([r * x, s * y], P)
        sage: g = DynamicalSystem([s * x, r * y], P)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over
         Multivariate Polynomial Ring in r, s over Rational Field
         defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over
           Multivariate Polynomial Ring in r, s over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (r*x : s*y)
          Dynamical System of Projective Space of dimension 1 over
           Multivariate Polynomial Ring in r, s over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (s*x : r*y)

    A dynamical semigroup may contain dynamical systems over finite fields::

        sage: F = FiniteField(5)
        sage: P.<x,y> = ProjectiveSpace(F, 1)
        sage: DynamicalSemigroup(([x, y], [x^2, y^2]))
        Dynamical semigroup over Projective Space of dimension 1 over
         Finite Field of size 5 defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over Finite Field of size 5
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over Finite Field of size 5
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    If a dynamical semigroup is built from dynamical systems over both projective and
    affine spaces, all systems will be homogenized to dynamical systems over projective space::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: A.<z> = AffineSpace(QQ, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem(z^2, A)
        sage: DynamicalSemigroup((f, g))
        Dynamical semigroup over Projective Space of dimension 1 over Rational Field
         defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)

    TESTS::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: DynamicalSemigroup(1)
        Traceback (most recent call last):
        ...
        TypeError: 1 does not define a 'DynamicalSemigroup' object

    ::

        sage: # needs sage.rings.number_field
        sage: R.<r> = QQ[]
        sage: K.<k> = NumberField(r^2 - 2)
        sage: P.<x,y> = ProjectiveSpace(RR, 1)
        sage: Q.<z,w> = ProjectiveSpace(K, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([z^2, w^2], Q)
        sage: DynamicalSemigroup((f, g))
        Traceback (most recent call last):
        ...
        ValueError: given dynamical systems are not automorphic under global composition

    ::

        sage: F = FiniteField(5)
        sage: P.<x,y> = ProjectiveSpace(F, 1)
        sage: Q.<z,w> = ProjectiveSpace(QQ, 1)
        sage: f = DynamicalSystem([x, y], P)
        sage: g = DynamicalSystem([z^2, w^2], Q)
        sage: DynamicalSemigroup((f, g))
        Traceback (most recent call last):
        ...
        ValueError: given dynamical systems are not automorphic under global composition

    ::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: Q.<u,v,w> = ProjectiveSpace(QQ, 2)
        sage: f = DynamicalSystem([x, y])
        sage: g = DynamicalSystem([u^2, v^2, w^2])
        sage: DynamicalSemigroup((f, g))
        Traceback (most recent call last):
        ...
        ValueError: domains of 'DynamicalSystem' objects must be of the same dimension
    """

    @staticmethod
    def __classcall_private__(cls, ds_data):
        if isinstance(ds_data, Collection):
            all_affine_systems = all(isinstance(ds_datum, DynamicalSystem_affine) for ds_datum in ds_data)
            if all_affine_systems:
                return DynamicalSemigroup_affine(ds_data)
        elif isinstance(ds_data, DynamicalSystem_affine):
            return DynamicalSemigroup_affine(ds_data)
        elif not isinstance(ds_data, DynamicalSystem):
            raise TypeError(str(ds_data) + " does not define a 'DynamicalSemigroup' object")
        return DynamicalSemigroup_projective(ds_data)

    def __init__(self, systems):
        r"""
        The Python constructor.

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: DynamicalSemigroup(([x, y], [x^2, y^2]))
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field defined by 2 dynamical systems:
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x : y)
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^2 : y^2)
        """

        self._dynamical_systems = []
        for ds in systems:
            if ds not in self._dynamical_systems:
                self._dynamical_systems.append(ds)
        Parent.__init__(self, category=Semigroups().FinitelyGeneratedAsMagma())

    def __call__(self, input):
        r"""
        The result after evaluating this dynamical semigroup on a value.

        INPUT:

        - ``input`` -- one value that can be evaluated
          with the generators of this dynamical semigroup

        OUTPUT: a set of the resulting values after applying all of this
        dynamical semigroup's generators to ``input``

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f(2)
            {(2 : 1), (4 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f([2, 1])
            {(2 : 1), (4 : 1)}

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f(f(2))
            Traceback (most recent call last):
            ...
            TypeError: unable to convert {(4 : 1), (2 : 1)} to an element of Rational Field
        """
        result = set()
        for ds in self.defining_systems():
            result.add(ds(self.domain()(input)))
        return result

    def base_ring(self):
        r"""
        The base ring of this dynamical semigroup. This is identical
        to the base ring of all of its defining dynamical system.

        OUTPUT: a ring

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.base_ring()
            Rational Field
        """
        return self.defining_systems()[0].base_ring()

    def change_ring(self, new_ring):
        r"""
        Return a new :class:`DynamicalSemigroup` whose generators
        are the initial dynamical systems coerced to ``new_ring``.

        INPUT:

        - ``new_ring`` -- a ring

        OUTPUT:

        A :class:`DynamicalSemigroup` defined by this dynamical
        semigroup's generators, but coerced to ``new_ring``.

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.change_ring(RR)
            Dynamical semigroup over Projective Space of dimension 1 over
             Real Field with 53 bits of precision defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over
               Real Field with 53 bits of precision
                Defn: Defined on coordinates by sending (x : y) to (x : y)
              Dynamical System of Projective Space of dimension 1 over
               Real Field with 53 bits of precision
                Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)
        """
        new_systems = []
        for ds in self.defining_systems():
            new_systems.append(ds.change_ring(new_ring))
        return DynamicalSemigroup_projective(new_systems)

    def domain(self):
        r"""
        Return the domain of the generators of this dynamical semigroup.

        OUTPUT: a subscheme of a projective space or affine space

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.domain()
            Projective Space of dimension 1 over Rational Field
        """
        return self.defining_systems()[0].domain()

    def codomain(self):
        r"""
        Return the codomain of the generators of this dynamical semigroup.

        OUTPUT: a subscheme of a projective space or affine space

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.codomain()
            Projective Space of dimension 1 over Rational Field
        """
        return self.defining_systems()[0].codomain()

    def defining_polynomials(self):
        r"""
        Return the set of polynomials that define the generators of this dynamical semigroup.

        OUTPUT: a set of polynomials

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.defining_polynomials()
            {(x, y), (x^2, y^2)}
        """
        result = set()
        for ds in self.defining_systems():
            result.add(ds.defining_polynomials())
        return result

    def defining_systems(self):
        r"""
        Return the generators of this dynamical semigroup.

        OUTPUT: a tuple of dynamical systems

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f.defining_systems()
            (Dynamical System of Projective Space of dimension 1 over Rational Field
               Defn: Defined on coordinates by sending (x : y) to (x : y),
             Dynamical System of Projective Space of dimension 1 over Rational Field
               Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2))
        """
        return tuple(self._dynamical_systems)

    def nth_iterate(self, p, n):
        r"""
        Return a set of values that results from evaluating this dynamical semigroup
        on the value ``p`` a total of ``n`` times.

        INPUT:

        - ``p`` -- a value on which dynamical systems can evaluate
        - ``n`` -- nonnegative integer

        OUTPUT: a set of values

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^2, y^2],))
            sage: f.nth_iterate(2, 0)
            {(2 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^2, y^2],))
            sage: f.nth_iterate(2, 1)
            {(4 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^2, y^2],))
            sage: f.nth_iterate(2, 2)
            {(16 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(2, 0)
            {(2 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(2, 1)
            {(3 : 1), (4 : 1)}

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(2, 2)
            {(5/3 : 1), (2 : 1), (9 : 1), (16 : 1)}

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(2, 3.5)
            Traceback (most recent call last):
            ...
            TypeError: Attempt to coerce non-integral RealNumber to Integer

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(2, -3)
            Traceback (most recent call last):
            ...
            ValueError: -3 must be a nonnegative integer

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: f.nth_iterate(3, 2) == (f * f)(3)
            True

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x + y, x - y], [x^2, y^2]))
            sage: one = QQ(1)
            sage: f.nth_iterate(2, one)
            {(3 : 1), (4 : 1)}
        """
        n = ZZ(n)
        if n < 0:
            raise ValueError(str(n) + " must be a nonnegative integer")
        result = {self.domain()(p)}
        for i in range(1, n + 1):
            next_iteration = set()
            for point in result:
                next_iteration.update(self(point))
            result = next_iteration
        return result

    def orbit(self, p, n):
        r"""
        If ``n`` is an integer, return `(p, f(p), f^2(p), \dots, f^n(p))`. If ``n`` is a list or tuple in interval
        notation `[a, b]`, return `(f^a(p), \dots, f^b(p))`.

        INPUT:

        - ``p`` -- value on which this dynamical semigroup can be evaluated
        - ``n`` -- nonnegative integer or a list or tuple of length 2 describing an
          interval of the number line containing entirely nonnegative integers

        OUTPUT: a tuple of sets of values on the domain of this dynamical semigroup

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, 0)
            ({(2 : 1)},)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, 1)
            ({(2 : 1)}, {(2 : 1), (4 : 1)})

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, 2)
            ({(2 : 1)}, {(2 : 1), (4 : 1)}, {(2 : 1), (4 : 1), (16 : 1)})

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, [1, 2])
            ({(2 : 1), (4 : 1)}, {(2 : 1), (4 : 1), (16 : 1)})

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: one = QQ(1)
            sage: d.orbit(2, one)
            ({(2 : 1)}, {(2 : 1), (4 : 1)})

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, -2)
            Traceback (most recent call last):
            ...
            ValueError: -2 must be a nonnegative integer

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, x)
            Traceback (most recent call last):
            ...
            TypeError: x is not a constant polynomial

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, [1, 2, 3])
            Traceback (most recent call last):
            ...
            ValueError: [1, 2, 3] must be an integer or list or tuple of two integers

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, [-2, 1])
            Traceback (most recent call last):
            ...
            ValueError: [-2, 1] must contain exactly two nonnegative integers

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: d = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: d.orbit(2, [2, 1])
            Traceback (most recent call last):
            ...
            ValueError: [2, 1] cannot be in descending order
        """

        if not isinstance(n, Collection):
            n = ZZ(n)
            if n < 0:
                raise ValueError(str(n) + " must be a nonnegative integer")
            return self.orbit(p, [0, n])

        if not len(n) == 2:
            raise ValueError(str(n) + " must be an integer or list or tuple of two integers")
        if ZZ(n[0]) < 0 or ZZ(n[1]) < 0:
            raise ValueError(str(n) + " must contain exactly two nonnegative integers")
        if ZZ(n[0]) > ZZ(n[1]):
            raise ValueError(str(n) + " cannot be in descending order")

        result = []
        current_iterate = self.nth_iterate(p, n[0])
        result.append(current_iterate)
        for i in range(n[0] + 1, n[1] + 1):
            next_iterate = set()
            for value in current_iterate:
                next_iterate.update(self(value))
            result.append(next_iterate)
            current_iterate = next_iterate
        return tuple(result)

    def specialization(self, assignments):
        r"""
        Return the specialization of the generators of this dynamical semigroup.

        INPUT:

        - ``assignments`` -- argument for specialization of the generators of
          this dynamical semigroup

        OUTPUT: a dynamical semigroup with the specialization of the generators
        of this dynamical semigroup

        EXAMPLES::

            sage: R.<r> = QQ[]
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: f = DynamicalSystem([r * x, y], P)
            sage: g = DynamicalSystem([x, r * y], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.specialization({r:2})
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (2*x : y)
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (x : 2*y)

        ::

            sage: R.<r> = QQ[]
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: f = DynamicalSystem([r * x, y], P)
            sage: g = DynamicalSystem([x, y], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.specialization({r:2})
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (2*x : y)
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (x : y)

        ::

            sage: R.<r,s> = QQ[]
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: f = DynamicalSystem([r * x, y], P)
            sage: g = DynamicalSystem([s * x, y], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.specialization({r:2, s:3})
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (2*x : y)
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (3*x : y)

        ::

            sage: R.<r,s> = QQ[]
            sage: P.<x,y> = ProjectiveSpace(R, 1)
            sage: f = DynamicalSystem([r * x, s * y], P)
            sage: g = DynamicalSystem([s * x, r * y], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.specialization({s:3})
            Dynamical semigroup over Projective Space of dimension 1 over
             Univariate Polynomial Ring in r over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over
               Univariate Polynomial Ring in r over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (r*x : 3*y)
              Dynamical System of Projective Space of dimension 1 over
               Univariate Polynomial Ring in r over Rational Field
                Defn: Defined on coordinates by sending (x : y) to (3*x : r*y)
        """
        specialized_systems = []
        for ds in self.defining_systems():
            specialized_systems.append(ds.specialization(assignments))
        return DynamicalSemigroup(specialized_systems)

    def __mul__(self, other_dynamical_semigroup):
        r"""
        Return a new :class:`DynamicalSemigroup` that is the result of multiplying
        this dynamical semigroup with another dynamical semigroup of the same type
        using the * operator.

        Let `f` be a dynamical semigroup with generators `\{ f_1, f_2, \dots, f_m \}`
        and `g` be a dynamical semigroup with generators `\{ g_1, g_2, \dots, g_n \}`.
        The product `f * g` has generators
        `\{ f_i(g_j) : 1 \leq i \leq m, 1 \leq j \leq n \}`.

        INPUT:

        - ``other_dynamical_semigroup`` -- a dynamical semigroup

        OUTPUT: :class:`DynamicalSemigroup`

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x^10, y^10], P)
            sage: f = DynamicalSemigroup(f1)
            sage: f*f
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field defined by 1 dynamical system:
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^100 : y^100)

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem_affine(x^10, A)
            sage: f = DynamicalSemigroup(f1)
            sage: f*f
            Dynamical semigroup over Affine Space of dimension 1 over Rational Field defined by 1 dynamical system:
            Dynamical System of Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x) to
                    (x^100)

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x^2 + x * y, y^2 + x * y], P)
            sage: g1 = DynamicalSystem([x^3 + x^2 * y, y^3 + x * y^2], P)
            sage: f = DynamicalSemigroup(f1)
            sage: g = DynamicalSemigroup(g1)
            sage: f*g
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field defined by 1 dynamical system:
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^6 + 2*x^5*y + 2*x^4*y^2 + 2*x^3*y^3 + x^2*y^4 : x^4*y^2 + 2*x^3*y^3 + 2*x^2*y^4 + 2*x*y^5 + y^6)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x^2 + x * y, y^2 + x * y], P)
            sage: f2 = DynamicalSystem([x^2 - x * y, y^2 - x * y], P)
            sage: g1 = DynamicalSystem([x^3 + x^2 * y, y^3 + x * y^2], P)
            sage: g2 = DynamicalSystem([x^3 - x^2 * y, y^3 - x * y^2], P)
            sage: f = DynamicalSemigroup((f1, f2))
            sage: g = DynamicalSemigroup((g1, g2))
            sage: f*g
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field defined by 2 dynamical systems:
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^6 + 2*x^5*y + 2*x^4*y^2 + 2*x^3*y^3 + x^2*y^4 : x^4*y^2 + 2*x^3*y^3 + 2*x^2*y^4 + 2*x*y^5 + y^6)
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^6 - 2*x^5*y + 2*x^3*y^3 - x^2*y^4 : -x^4*y^2 + 2*x^3*y^3 - 2*x*y^5 + y^6)

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x^2 + x * y, y^2 - x * y], P)
            sage: g1 = DynamicalSystem([x^3 - x^2 * y, y^3 + x * y^2], P)
            sage: f = DynamicalSemigroup(f1)
            sage: g = DynamicalSemigroup(g1)
            sage: f*g == g*f
            False

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x^2 + x * y, y^2 + x * y], P)
            sage: g1 = DynamicalSystem([x^3 + x^2 * y, y^3 + x * y^2], P)
            sage: h1 = DynamicalSystem([x^4 + x^3 * y, y^4 + x * y^3], P)
            sage: f = DynamicalSemigroup(f1)
            sage: g = DynamicalSemigroup(g1)
            sage: h = DynamicalSemigroup(h1)
            sage: (f*g)*h == f*(g*h)
            True

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x, y], P)
            sage: f = DynamicalSemigroup(f1)
            sage: f*1
            Traceback (most recent call last):
            ...
            TypeError: can only multiply dynamical semigroups with other dynamical semigroups of the same type

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: A.<z> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem_projective([x, y], P)
            sage: g1 = DynamicalSystem_affine(z, A)
            sage: f = DynamicalSemigroup_projective(f1)
            sage: g = DynamicalSemigroup_affine(g1)
            sage: f*g
            Traceback (most recent call last):
            ...
            TypeError: can only multiply dynamical semigroups with other dynamical semigroups of the same type

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: Q.<z,w> = ProjectiveSpace(RR, 1)
            sage: f1 = DynamicalSystem([x, y], P)
            sage: g1 = DynamicalSystem([z, w], Q)
            sage: f = DynamicalSemigroup(f1)
            sage: g = DynamicalSemigroup(g1)
            sage: f*g
            Traceback (most recent call last):
            ...
            ValueError: left dynamical semigroup's domain must equal right dynamical semigroup's codomain

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: Q.<z,w> = ProjectiveSpace(QQ, 1)
            sage: f1 = DynamicalSystem([x, y], P)
            sage: g1 = DynamicalSystem([z, w], Q)
            sage: f = DynamicalSemigroup(f1)
            sage: g = DynamicalSemigroup(g1)
            sage: f*g
            Traceback (most recent call last):
            ...
            ValueError: left dynamical semigroup's domain must equal right dynamical semigroup's codomain
        """
        if type(self) is not type(other_dynamical_semigroup):
            raise TypeError("can only multiply dynamical semigroups with other dynamical semigroups of the same type")
        if self.domain() != other_dynamical_semigroup.codomain():
            raise ValueError("left dynamical semigroup's domain must equal right dynamical semigroup's codomain")
        composite_systems = []
        for f in self.defining_systems():
            for g in other_dynamical_semigroup.defining_systems():
                composite_systems.append(DynamicalSystem(f * g))
        return DynamicalSemigroup(composite_systems)

    def __pow__(self, n):
        r"""
        Return a new dynamical semigroup that is this dynamical semigroup with itself ``n`` times.
        If ``n`` is zero, return a dynamical semigroup with the identity map.

        INPUT:

        - ``n`` -- nonnegative integer

        OUTPUT: :class:`DynamicalSemigroup`

        EXAMPLES::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f = DynamicalSystem(x^2, A)
            sage: d = DynamicalSemigroup(f)
            sage: d^2
            Dynamical semigroup over Affine Space of dimension 1 over Rational Field defined by 1 dynamical system:
            Dynamical System of Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x) to
                    (x^4)

        ::

            sage: A.<x, y, z, w, t, u ,v> = AffineSpace(QQ, 7)
            sage: f = DynamicalSystem([t + u, v - w, x + y, z^2, u * t, v^2 - w^2, x * y * z], A)
            sage: d = DynamicalSemigroup(f)
            sage: d^0
            Dynamical semigroup over Affine Space of dimension 7 over Rational Field defined by 1 dynamical system:
            Dynamical System of Affine Space of dimension 7 over Rational Field
              Defn: Defined on coordinates by sending (x, y, z, w, t, u, v) to
                    (x, y, z, w, t, u, v)

        TESTS::

            sage: A.<x, y, z, w, t, u ,v> = AffineSpace(QQ, 7)
            sage: f = DynamicalSystem([t + u, v - w, x + y, z^2, u * t, v^2 - w^2, x * y * z], A)
            sage: d = DynamicalSemigroup(f)
            sage: d*d == d^2
            True
            sage: d^3 * d^2 == d^(3 + 2)
            True
            sage: (d^3)^2 == d^(3 * 2)
            True

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem(x^2 + 1, A)
            sage: f2 = DynamicalSystem(x^3 - 1, A)
            sage: g1 = DynamicalSystem(x^3 + x - 1, A)
            sage: g2 = DynamicalSystem(x^2 - x + 1, A)
            sage: f = DynamicalSemigroup((f1, f2))
            sage: g = DynamicalSemigroup((g1, g2))
            sage: (f * g) ^ 2 == f^2 * g^2
            False

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem(x^2 + 1, A)
            sage: f2 = DynamicalSystem(x^3 - 1, A)
            sage: g1 = DynamicalSystem(x^3 + x - 1, A)
            sage: g2 = DynamicalSystem(x^2 - x + 1, A)
            sage: f = DynamicalSemigroup((f1, f2))
            sage: g = DynamicalSemigroup((g1, g2))
            sage: f * g^0 == f
            True

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem(x^2 + 1, A)
            sage: f2 = DynamicalSystem(x^3 - 1, A)
            sage: g1 = DynamicalSystem(x^3 + x - 1, A)
            sage: g2 = DynamicalSystem(x^2 - x + 1, A)
            sage: f = DynamicalSemigroup((f1, f2))
            sage: g = DynamicalSemigroup((g1, g2))
            sage: g * f^0 == g
            True

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem(x^2 + 1, A)
            sage: f2 = DynamicalSystem(x^3 - 1, A)
            sage: d = DynamicalSemigroup((f1, f2))
            sage: d^1.5
            Traceback (most recent call last):
            ...
            TypeError: Attempt to coerce non-integral RealNumber to Integer

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f1 = DynamicalSystem(x^2 + 1, A)
            sage: f2 = DynamicalSystem(x^3 - 1, A)
            sage: d = DynamicalSemigroup((f1, f2))
            sage: d^(-1)
            Traceback (most recent call last):
            ...
            ValueError: -1 must be a nonnegative integer

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f = DynamicalSystem(x^2, A)
            sage: d = DynamicalSemigroup(f)
            sage: two = RR(2)
            sage: d^two
            Dynamical semigroup over Affine Space of dimension 1 over Rational Field defined by 1 dynamical system:
            Dynamical System of Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x) to
                    (x^4)
        """
        n = ZZ(n)
        if n < 0:
            raise ValueError(str(n) + " must be a nonnegative integer")
        if n == 0:
            return DynamicalSemigroup(DynamicalSystem(self.defining_systems()[0] ** 0))
        result = self
        for i in range(n - 1):
            result = result * self
        return result

    def _repr_(self):
        r"""
        Return the :class:`String` representation of this dynamical semigroup.

        OUTPUT: a :class:`String` displaying information about this dynamical semigroup

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: DynamicalSemigroup(([x, y], [x^2, y^2]))
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field defined by 2 dynamical systems:
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x : y)
            Dynamical System of Projective Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x : y) to
                    (x^2 : y^2)
        """
        header = "Dynamical semigroup over %s defined by %d dynamical system"
        if (len(self.defining_systems()) > 1):
            header += "s"
        header += ":"
        header = header % (str(self.domain()), len(self.defining_systems()))
        systems = []
        for ds in self.defining_systems():
            systems.append(str(ds))
        systems = '\n'.join(systems)
        return header + "\n" + systems

    def __eq__(self, other):
        r"""
        Return whether two dynamical semigroups are equal.

        OUTPUT:

        A boolean that is ``True`` if and only if the generators of the two
        dynamical semigroups are equal as sets and no generator is of degree 1.

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^2, y^2], [x^3, y^3]))
            sage: g = DynamicalSemigroup(([x^2, y^2], [x^3, y^3]))
            sage: f == g
            True

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^2, y^2], [x^2, y^2]))
            sage: g = DynamicalSemigroup(([x^2, y^2], [x^2, y^2], [x^2, y^2]))
            sage: f == g
            True

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^3, y^3], [x^2, y^2]))
            sage: g = DynamicalSemigroup(([x^2, y^2], [x^3, y^3]))
            sage: f == g
            True

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x^3, y^3], [x^2, y^2]))
            sage: g = DynamicalSemigroup(([x^2, y^2], [y^3, x^3]))
            sage: f == g
            False

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: f == 1
            False

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSemigroup(([x, y], [x^2, y^2]))
            sage: g = DynamicalSemigroup(([x^2, y^2], [x^3, y^3]))
            sage: f == g
            Traceback (most recent call last):
            ...
            NotImplementedError: cannot compare dynamical semigroups with at least one generator of degree 1
        """
        if isinstance(other, DynamicalSemigroup):
            if any(ds.degree() == 1 for ds in self.defining_systems()) or \
                    any(ds.degree() == 1 for ds in other.defining_systems()):
                raise NotImplementedError("cannot compare dynamical semigroups with at least one generator of degree 1")
            return all(ds in other.defining_systems() for ds in self.defining_systems()) and \
                all(ds in self.defining_systems() for ds in other.defining_systems())
        return False


class DynamicalSemigroup_projective(DynamicalSemigroup):
    r"""
    A dynamical semigroup defined by a multiple dynamical systems on projective space.

    INPUT:

    - ``ds_data`` -- list or tuple of dynamical systems or objects that define
      dynamical systems over projective space

    OUTPUT: :class:`DynamicalSemigroup_projective`

    EXAMPLES::

        sage: P.<x,y> = ProjectiveSpace(QQ, 1)
        sage: DynamicalSemigroup_projective(([x, y], [x^2, y^2]))
        Dynamical semigroup over Projective Space of dimension 1 over
         Rational Field defined by 2 dynamical systems:
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x : y)
          Dynamical System of Projective Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x : y) to (x^2 : y^2)
    """

    @staticmethod
    def __classcall_private__(cls, ds_data):
        systems = []

        if isinstance(ds_data, Collection):
            for ds_datum in ds_data:
                if isinstance(ds_datum, DynamicalSystem_projective):
                    systems.append(ds_datum)
                elif isinstance(ds_datum, DynamicalSystem_affine):
                    dimension = ds_datum.domain().dimension()
                    systems.append(ds_datum.homogenize(dimension))
                else:
                    try:
                        systems.append(DynamicalSystem_projective(ds_datum))
                    except ValueError:
                        raise ValueError(str(ds_datum) + " does not define a 'DynamicalSystem_projective' object")
        else:
            if isinstance(ds_data, DynamicalSystem_projective):
                systems.append(ds_data)
            else:
                try:
                    systems.append(DynamicalSystem_projective(ds_data))
                except ValueError:
                    raise ValueError(str(ds_data) + " does not define a 'DynamicalSystem_projective' object")

        systems = _standardize_domains_of_(systems)
        if systems[0].base_ring() not in Fields():
            return typecall(cls, systems)
        if isinstance(systems[0].base_ring(), FiniteField):
            return DynamicalSemigroup_projective_finite_field(systems)
        return DynamicalSemigroup_projective_field(systems)

    def dehomogenize(self, n):
        r"""
        Return a new :class:`DynamicalSemigroup_projective` with the dehomogenization at ``n`` of
        the generators of this dynamical semigroup.

        INPUT:

        - ``n`` -- tuple of nonnegative integers; if `n` is an integer,
          then the two values of the tuple are assumed to be the same

        OUTPUT: :class:`DynamicalSemigroup_affine`

        EXAMPLES::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSystem([x, y], P)
            sage: g = DynamicalSystem([x^2, y^2], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.dehomogenize(0)
            Dynamical semigroup over Affine Space of dimension 1 over
             Rational Field defined by 2 dynamical systems:
              Dynamical System of Affine Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (y) to (y)
              Dynamical System of Affine Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (y) to (y^2)

        ::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSystem([x, y], P)
            sage: g = DynamicalSystem([x^2, y^2], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.dehomogenize(1)
            Dynamical semigroup over Affine Space of dimension 1 over
             Rational Field defined by 2 dynamical systems:
              Dynamical System of Affine Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x) to (x)
              Dynamical System of Affine Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x) to (x^2)

        TESTS::

            sage: P.<x,y> = ProjectiveSpace(QQ, 1)
            sage: f = DynamicalSystem([x, y], P)
            sage: g = DynamicalSystem([x^2, y^2], P)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.dehomogenize((1, 0))
            Traceback (most recent call last):
            ...
            ValueError: Scheme morphism:
              From: Affine Space of dimension 1 over Rational Field
              To:   Affine Space of dimension 1 over Rational Field
              Defn: Defined on coordinates by sending (x) to (1/x)
            is not a `DynamicalSystem_affine` object
        """
        new_systems = []
        for ds in self.defining_systems():
            new_system = ds.dehomogenize(n)
            if not isinstance(new_system, DynamicalSystem_affine):
                raise ValueError(str(new_system) + " is not a `DynamicalSystem_affine` object")
            new_systems.append(new_system)
        return DynamicalSemigroup_affine(new_systems)


class DynamicalSemigroup_projective_field(DynamicalSemigroup_projective):
    pass


class DynamicalSemigroup_projective_finite_field(DynamicalSemigroup_projective_field):
    pass


class DynamicalSemigroup_affine(DynamicalSemigroup):
    r"""
    A dynamical semigroup defined by multiple dynamical systems on affine space.

    INPUT:

    - ``ds_data`` -- list or tuple of dynamical systems or objects that define dynamical systems
      over affine space

    OUTPUT: :class:`DynamicalSemigroup_affine`

    EXAMPLES::

        sage: A.<x> = AffineSpace(QQ, 1)
        sage: f = DynamicalSystem(x, A)
        sage: g = DynamicalSystem(x^2, A)
        sage: DynamicalSemigroup_affine((f, g))
        Dynamical semigroup over Affine Space of dimension 1 over
         Rational Field defined by 2 dynamical systems:
          Dynamical System of Affine Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x) to (x)
          Dynamical System of Affine Space of dimension 1 over Rational Field
            Defn: Defined on coordinates by sending (x) to (x^2)
    """

    @staticmethod
    def __classcall_private__(cls, ds_data):
        systems = []

        if isinstance(ds_data, Collection):
            for ds_datum in ds_data:
                if isinstance(ds_datum, DynamicalSystem_affine):
                    systems.append(ds_datum)
                else:
                    try:
                        systems.append(DynamicalSystem_affine(ds_datum))
                    except ValueError:
                        raise ValueError(str(ds_datum) + " does not define a 'DynamicalSystem_affine' object")
        else:
            if isinstance(ds_data, DynamicalSystem_affine):
                systems.append(ds_data)
            else:
                try:
                    systems.append(DynamicalSystem_affine(ds_data))
                except ValueError:
                    raise ValueError(str(ds_data) + " does not define a 'DynamicalSystem_affine' object")

        systems = _standardize_domains_of_(systems)
        if systems[0].base_ring() not in Fields():
            return typecall(cls, systems)
        if isinstance(systems[0].base_ring(), FiniteField):
            return DynamicalSemigroup_affine_finite_field(systems)
        return DynamicalSemigroup_affine_field(systems)

    def homogenize(self, n):
        r"""
        Return a new :class:`DynamicalSemigroup_projective` with the homogenization at ``n`` of
        the generators of this dynamical semigroup.

        INPUT:

        - ``n`` -- tuple of nonnegative integers; if `n` is an integer,
          then the two values of the tuple are assumed to be the same

        OUTPUT: :class:`DynamicalSemigroup_projective`

        EXAMPLES::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f = DynamicalSystem(x + 1, A)
            sage: g = DynamicalSystem(x^2, A)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.homogenize(1)
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x0 : x1) to (x0 + x1 : x1)
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x0 : x1) to (x0^2 : x1^2)

        ::

            sage: A.<x> = AffineSpace(QQ, 1)
            sage: f = DynamicalSystem(x + 1, A)
            sage: g = DynamicalSystem(x^2, A)
            sage: d = DynamicalSemigroup((f, g))
            sage: d.homogenize((1, 0))
            Dynamical semigroup over Projective Space of dimension 1 over Rational Field
             defined by 2 dynamical systems:
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x0 : x1) to (x1 : x0 + x1)
              Dynamical System of Projective Space of dimension 1 over Rational Field
                Defn: Defined on coordinates by sending (x0 : x1) to (x1^2 : x0^2)
        """
        new_systems = []
        for ds in self.defining_systems():
            new_systems.append(ds.homogenize(n))
        return DynamicalSemigroup_projective(new_systems)


class DynamicalSemigroup_affine_field(DynamicalSemigroup_affine):
    pass


class DynamicalSemigroup_affine_finite_field(DynamicalSemigroup_affine_field):
    pass


def _standardize_domains_of_(systems):
    r"""
    Coerces dynamical systems to the same domain and have the same generators.

    INPUT:

    - ``systems`` -- list of dynamical systems

    OUTPUT: list of dynamical systems from ``systems`` coerced to the same domain with the same generators

    EXAMPLES::

        sage: A.<x> = AffineSpace(QQ, 1)
        sage: B.<y> = AffineSpace(RR, 1)
        sage: f = DynamicalSystem(x, A)
        sage: g = DynamicalSystem(x^2, B)
        sage: sage.dynamics.arithmetic_dynamics.dynamical_semigroup._standardize_domains_of_([f, g])
        [Dynamical System of Affine Space of dimension 1 over Real Field with 53 bits of precision
           Defn: Defined on coordinates by sending (x) to (x),
         Dynamical System of Affine Space of dimension 1 over Real Field with 53 bits of precision
           Defn: Defined on coordinates by sending (x) to (x^2)]
    """
    identical_domains = True
    for ds in systems:
        if ds.domain() != systems[0].domain():
            identical_domains = False
            break

    over_number_fields = True
    all_over_QQ = True
    for ds in systems:
        if ds.base_ring() not in NumberFields():
            over_number_fields = False
        if ds.base_ring() is not QQ:
            all_over_QQ = False

    biggest_ring = None

    if over_number_fields and not all_over_QQ:
        number_fields = []
        for ds in systems:
            number_fields.append(ds.base_ring())

        minimal_composite_field = None
        for field in number_fields:
            if field is not QQ:
                if minimal_composite_field is None:
                    minimal_composite_field = field
                else:
                    minimal_composite_field = minimal_composite_field.composite_fields(field)[0]

        biggest_ring = minimal_composite_field
    else:
        for ds in systems:
            if biggest_ring is None:
                biggest_ring = ds.base_ring()
            elif ds.base_ring().has_coerce_map_from(biggest_ring):
                biggest_ring = ds.base_ring()
            elif biggest_ring.has_coerce_map_from(ds.base_ring()):
                pass
            else:
                raise ValueError("given dynamical systems are not automorphic \
                                under global composition")

    for i in range(len(systems)):
        if systems[i].base_ring() != biggest_ring:
            systems[i] = systems[i].change_ring(biggest_ring)

    domain = systems[0].domain()

    identical_domains = all(ds.domain() == systems[0].domain() for ds in systems)
    if not identical_domains:
        for ds in systems:
            if ds.domain().dimension() != systems[0].domain().dimension():
                raise ValueError("domains of 'DynamicalSystem' objects must be of the same dimension")

    gens = systems[0].domain().ambient_space().gens()
    for i in range(len(systems)):
        if systems[i].domain().coordinate_ring() != systems[0].domain().coordinate_ring():
            sub_dict = {}
            old_gens = systems[i].domain().ambient_space().gens()
            for j in range(len(old_gens)):
                sub_dict[old_gens[j]] = gens[j]
            new_polys = []
            for poly in systems[i].defining_polynomials():
                new_polys.append(poly.subs(sub_dict))
            systems[i] = DynamicalSystem(new_polys, domain)

    return systems



"""
Root system data for (untwisted) type F affine
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Daniel Bump
#       Copyright (C) 2008-2009 Justin Walker
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .cartan_type import CartanType_standard_untwisted_affine


class CartanType(CartanType_standard_untwisted_affine):
    def __init__(self):
        """
        EXAMPLES::

            sage: ct = CartanType(['F',4,1])
            sage: ct
            ['F', 4, 1]
            sage: ct._repr_(compact = True)
            'F4~'

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            False
            sage: ct.is_affine()
            True
            sage: ct.is_untwisted_affine()
            True
            sage: ct.is_crystallographic()
            True
            sage: ct.is_simply_laced()
            False
            sage: ct.classical()
            ['F', 4]
            sage: ct.dual()
            ['F', 4, 1]^*
            sage: ct.dual().is_untwisted_affine()
            False

        TESTS::

            sage: TestSuite(ct).run()
        """
        CartanType_standard_untwisted_affine.__init__(self, "F", 4)

    def dynkin_diagram(self):
        """
        Return the extended Dynkin diagram for affine type F.

        EXAMPLES::

            sage: f = CartanType(['F', 4, 1]).dynkin_diagram(); f                       # needs sage.graphs
            O---O---O=>=O---O
            0   1   2   3   4
            F4~
            sage: f.edges(sort=True)                                                    # needs sage.graphs
            [(0, 1, 1), (1, 0, 1), (1, 2, 1), (2, 1, 1),
             (2, 3, 2), (3, 2, 1), (3, 4, 1), (4, 3, 1)]
        """
        from .dynkin_diagram import DynkinDiagram_class
        g = DynkinDiagram_class(self)
        for i in range(1, 4):
            g.add_edge(i, i+1)
        g.set_edge_label(2,3,2)
        g.add_edge(0, 1)
        return g

    def _latex_dynkin_diagram(self, label=None, node=None, node_dist=2, dual=False):
        r"""
        Return a latex representation of the Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['F',4,1])._latex_dynkin_diagram())
            \draw (0 cm,0) -- (2 cm,0);
            {
            \pgftransformxshift{2 cm}
            \draw (0 cm,0) -- (2 cm,0);
            \draw (2 cm, 0.1 cm) -- +(2 cm,0);
            \draw (2 cm, -0.1 cm) -- +(2 cm,0);
            \draw (4.0 cm,0) -- +(2 cm,0);
            \draw[shift={(3.2, 0)}, rotate=0] (135 : 0.45cm) -- (0,0) -- (-135 : 0.45cm);
            \draw[fill=white] (0 cm, 0 cm) circle (.25cm) node[below=4pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (6 cm, 0 cm) circle (.25cm) node[below=4pt]{$4$};
            }
            \draw[fill=white] (0 cm, 0 cm) circle (.25cm) node[below=4pt]{$0$};
            <BLANKLINE>
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._latex_draw_node
        ret = "\\draw (0 cm,0) -- (%s cm,0);\n" % node_dist
        ret += "{\n\\pgftransformxshift{%s cm}\n" % node_dist
        ret += self.classical()._latex_dynkin_diagram(label, node, node_dist, dual)
        ret += "}\n" + node(0, 0, label(0))
        return ret

    def ascii_art(self, label=None, node=None):
        """
        Return a ascii art representation of the extended Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['F',4,1]).ascii_art(label = lambda x: x+2))
            O---O---O=>=O---O
            2   3   4   5   6
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._ascii_art_node
        ret = "{}---{}---{}=>={}---{}\n".format(node(label(0)), node(label(1)),
                             node(label(2)), node(label(3)), node(label(4)))
        ret += ("{!s:4}"*5 + "\n").format(label(0), label(1), label(2), label(3), label(4))
        return ret

    def _default_folded_cartan_type(self):
        """
        Return the default folded Cartan type.

        EXAMPLES::

            sage: CartanType(['F', 4, 1])._default_folded_cartan_type()
            ['F', 4, 1] as a folding of ['E', 6, 1]
        """
        from sage.combinat.root_system.type_folded import CartanTypeFolded
        return CartanTypeFolded(self, ['E', 6, 1], [[0], [2], [4], [3, 5], [1, 6]])



"""
Morphisms of free modules

AUTHORS:
    - William Stein: initial version

    - Miguel Marco (2010-06-19): added eigenvalues, eigenvectors and minpoly functions


TESTS::

    sage: V = ZZ^2; f = V.hom([V.1, -2*V.0])
    sage: loads(dumps(f))
    Free module morphism defined by the matrix
    [ 0  1]
    [-2  0]
    Domain:   Ambient free module of rank 2 over the principal ideal domain ...
    Codomain: Ambient free module of rank 2 over the principal ideal domain ...
    sage: loads(dumps(f)) == f
    True
"""

##########################################################################
#       Copyright (C) 2009 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
##########################################################################

# A matrix morphism is a morphism that is defined by multiplication by a
# matrix.  Elements of domain must either have a method "vector()" that
# returns a vector that the defining matrix can hit from the left, or
# be coercible into vector space of appropriate dimension.

from sage.modules import free_module

from sage.categories.morphism import Morphism
from sage.modules import free_module_homspace, matrix_morphism
from sage.structure.richcmp import rich_to_bool, richcmp
from sage.structure.sequence import Sequence


class FreeModuleMorphism(matrix_morphism.MatrixMorphism):

    def __init__(self, parent, A, side='left'):
        """
        INPUT:

            - ``parent`` -- a homspace in a (sub) category of free modules

            - ``A`` -- matrix

            - ``side`` -- side of the vectors acted on by the matrix  (default: ``'left'``)

        EXAMPLES::

            sage: V = ZZ^3; W = span([[1,2,3], [-1,2,8]], ZZ)
            sage: phi = V.hom(matrix(ZZ, 3, [1..9]))
            sage: type(phi)
            <class 'sage.modules.free_module_morphism.FreeModuleMorphism'>
        """
        if not isinstance(parent, free_module_homspace.FreeModuleHomspace):
            raise TypeError("parent (=%s) must be a free module hom space" % parent)
        if isinstance(A, matrix_morphism.MatrixMorphism):
            A = A.matrix()
        A = parent._matrix_space(side)(A)
        matrix_morphism.MatrixMorphism.__init__(self, parent, A, side=side)

    def pushforward(self, x):
        """
        Compute the image of a sub-module of the domain.

        EXAMPLES::

            sage: V = QQ^3; W = span([[1,2,3], [-1,2,5/3]], QQ)
            sage: phi = V.hom(matrix(QQ, 3, [1..9]))
            sage: phi.rank()
            2
            sage: phi(V)   #indirect doctest
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [ 1  0 -1]
            [ 0  1  2]

        We compute the image of a submodule of a ZZ-module embedded in
        a rational vector space::

            sage: V = QQ^3; W = V.span_of_basis([[2,2,3], [-1,2,5/3]], ZZ)
            sage: phi = W.hom([W.0, W.0 - W.1]); phi
            Free module morphism defined by the matrix
            [ 1  0]
            [ 1 -1]...
            sage: phi(span([2*W.1], ZZ))
            Free module of degree 3 and rank 1 over Integer Ring
            Echelon basis matrix:
            [  6   0 8/3]
            sage: phi(2*W.1)
            (6, 0, 8/3)
        """
        if isinstance(x, free_module.FreeModule_generic):
            V = self.domain().submodule(x)
            return self.restrict_domain(V).image()
        raise TypeError("`pushforward` is only defined for submodules")

    def _repr_(self):
        r"""
        Return string representation of this morphism of free modules.

        EXAMPLES::

            sage: V = ZZ^3; W = span([[1,2,3], [-1,2,8]], ZZ)
            sage: phi = V.hom(matrix(ZZ, 3, [1..9]))
            sage: phi._repr_()
            'Free module morphism defined by the matrix\n[1 2 3]\n[4 5 6]\n[7 8 9]\nDomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring\nCodomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring'

            sage: V = ZZ^6
            sage: W = ZZ^4
            sage: m = matrix(QQ, [[1, 0, 0 ,0], [0]*4, [0]*4, [0]*4, [0]*4, [0]*4])
            sage: phi = V.hom(m, W)
            sage: rho = phi.restrict_codomain(W.span([W.0]))
            sage: rho
            Free module morphism defined by the matrix
            [1]
            [0]
            [0]
            [0]
            [0]
            [0]
            Domain: Ambient free module of rank 6 over the principal ideal domain Integer Ring
            Codomain: Free module of degree 4 and rank 1 over Integer Ring
            Echelon basis matrix:
            [1 0 0 0]

            sage: V = QQ^40
            sage: m = matrix(QQ, 40, 40, 1600)
            sage: phi = V.hom(m, V)
            sage: phi
            Vector space morphism represented by the matrix:
            40 x 40 dense matrix over Rational Field
            Domain:   Vector space of dimension 40 over Rational Field
            Codomain: Vector space of dimension 40 over Rational Field

        The representation displays which side of the vectors the matrix is acting::

            sage: V = ZZ^3
            sage: h = V.hom([V.1, V.2, V.0]); h
            Free module morphism defined by the matrix
            [0 1 0]
            [0 0 1]
            [1 0 0]
            Domain:   Ambient free module of rank 3 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring
            sage: h2 = V.hom([V.1, V.2, V.0], side='right'); h2
            Free module morphism defined as left-multiplication by the matrix
            [0 0 1]
            [1 0 0]
            [0 1 0]
            Domain:   Ambient free module of rank 3 over the principal ideal domain Integer Ring
            Codomain: Ambient free module of rank 3 over the principal ideal domain Integer Ring
        """
        r = "Free module morphism defined {}by the matrix\n{!r}\nDomain: {}\nCodomain: {}"
        act = ""
        if self.side() == "right":
            act = "as left-multiplication "
        return r.format(act, self.matrix(), self.domain(), self.codomain())

    def change_ring(self, R):
        """
        Change the ring over which this morphism is defined.

        This changes the ring of the domain, codomain, and underlying matrix.

        EXAMPLES::

            sage: V0 = span([[0,0,1],[0,2,0]], ZZ); V1 = span([[1/2,0],[0,2]], ZZ)
            sage: W = span([[1,0],[0,6]], ZZ)
            sage: h = V0.hom([-3*V1.0 - 3*V1.1, -3*V1.0 - 3*V1.1])
            sage: h.base_ring()
            Integer Ring
            sage: h
            Free module morphism defined by the matrix
            [-3 -3]
            [-3 -3]...
            sage: h.change_ring(QQ).base_ring()
            Rational Field
            sage: f = h.change_ring(QQ); f
            Vector space morphism represented by the matrix:
            [-3 -3]
            [-3 -3]
            Domain:   Vector space of degree 3 and dimension 2 over Rational Field
                      Basis matrix:
                      [0 1 0]
                      [0 0 1]
            Codomain: Vector space of degree 2 and dimension 2 over Rational Field
                      Basis matrix:
                      [1 0]
                      [0 1]
            sage: f = h.change_ring(GF(7)); f
            Vector space morphism represented by the matrix:
            [4 4]
            [4 4]
            Domain:   Vector space of degree 3 and dimension 2 over Finite Field of size 7
                      Basis matrix:
                      [0 1 0]
                      [0 0 1]
            Codomain: Vector space of degree 2 and dimension 2 over Finite Field of size 7
                      Basis matrix:
                      [1 0]
                      [0 1]
        """
        D = self.domain().change_ring(R)
        C = self.codomain().change_ring(R)
        A = self.matrix().change_ring(R)
        return D.hom(A, C, side=self.side())

    def inverse_image(self, V):
        """
        Given a submodule V of the codomain of self, return the
        inverse image of V under self, i.e., the biggest submodule of
        the domain of ``self`` that maps into V.

        EXAMPLES:

        We test computing inverse images over a field::

            sage: V = QQ^3; W = span([[1,2,3], [-1,2,5/3]], QQ)
            sage: phi = V.hom(matrix(QQ, 3, [1..9]))
            sage: phi.rank()
            2
            sage: I = phi.inverse_image(W); I
            Vector space of degree 3 and dimension 2 over Rational Field
            Basis matrix:
            [   1    0    0]
            [   0    1 -1/2]
            sage: phi(I.0) in W
            True
            sage: phi(I.1) in W
            True
            sage: W = phi.image()
            sage: phi.inverse_image(W) == V
            True

        We test computing inverse images between two spaces embedded in different
        ambient spaces.::

            sage: V0 = span([[0,0,1],[0,2,0]],ZZ); V1 = span([[1/2,0],[0,2]],ZZ)
            sage: W = span([[1,0],[0,6]],ZZ)
            sage: h = V0.hom([-3*V1.0 - 3*V1.1, -3*V1.0 - 3*V1.1])
            sage: h.inverse_image(W)
            Free module of degree 3 and rank 2 over Integer Ring
            Echelon basis matrix:
            [0 2 1]
            [0 0 2]
            sage: h(h.inverse_image(W)).is_submodule(W)
            True
            sage: h(h.inverse_image(W)).index_in(W)
            +Infinity
            sage: h(h.inverse_image(W))
            Free module of degree 2 and rank 1 over Integer Ring
            Echelon basis matrix:
            [ 3 12]


        We test computing inverse images over the integers::

            sage: V = QQ^3; W = V.span_of_basis([[2,2,3],[-1,2,5/3]], ZZ)
            sage: phi = W.hom([W.0, W.0 - W.1])
            sage: Z = W.span([2*W.1]); Z
            Free module of degree 3 and rank 1 over Integer Ring
            Echelon basis matrix:
            [    2    -4 -10/3]
            sage: Y = phi.inverse_image(Z); Y
            Free module of degree 3 and rank 1 over Integer Ring
            Echelon basis matrix:
            [  6   0 8/3]
            sage: phi(Y) == Z
            True

        We test that :issue:`24590` is resolved::

            sage: A = FreeQuadraticModule(ZZ, 1, matrix([2]))
            sage: f = A.Hom(A).an_element()
            sage: f.inverse_image(A)
            Free module of degree 1 and rank 1 over Integer Ring
            Echelon basis matrix:
            [1]

        We test that it respects the ``side``::

            sage: V = ZZ^2
            sage: m = matrix(2, [1, 1, 0, 1])
            sage: h = V.hom(m, side='right')
            sage: h
            Free module morphism defined as left-multiplication by the matrix
            [1 1]
            [0 1]...
            sage: SV = V.span([V.0])
            sage: h.inverse_image(SV)
            Free module of degree 2 and rank 1 over Integer Ring
            Echelon basis matrix:
            [1 0]
            sage: V.hom(m).inverse_image(SV)
            Free module of degree 2 and rank 1 over Integer Ring
            Echelon basis matrix:
            [ 1 -1]
        """
        if self.rank() == 0:
            # Special case -- if this is the 0 map, then the only possibility
            # for the inverse image is that it is the whole domain.
            return self.domain()

        R = self.base_ring()
        if self.side() == "left":
            A = self.matrix()
        else:
            A = self.matrix().transpose()

        # Replace the module V that we are going to pullback by a
        # submodule that is contained in the image of self, since our
        # plan is to lift all generators of V.
        V = self.image().intersection(V)
        # Write V in terms of the basis for the codomain.
        V = self.codomain().coordinate_module(V)
        B = V.basis_matrix()

        # Compute the kernel, which is contained in the inverse image.
        K = self.kernel()

        if R.is_field():
            # By solving, find lifts of each of the basis elements of V.
            # Each row of C gives a linear combination of the basis for the domain
            # that maps to one of the basis elements V.
            C = A.solve_left(B)

        else:
            if not hasattr(A, 'hermite_form'):
                raise NotImplementedError("base ring (%s) must have hermite_form algorithm in order to compute inverse image" % R)

            # 1. Compute H such that U*A = H = hnf(A) without zero
            # rows. What this "does" is find a basis for the image of
            # A and explicitly represents each element in this basis
            # as the image of some element of the domain (the rows of
            # U give these elements of the domain).
            H, U = A.hermite_form(transformation=True, include_zero_rows=False)

            # 2. Next we find the unique solution to the equation
            #    Y*H = B.  This writes each basis element of V in
            # terms of our image basis found in the previous step.
            Y = H.solve_left(B)

            # 3. Multiply Y by U then takes those same linear combinations
            # from step 2 above and lifts them to coefficients that define
            # linear combinations of the basis for the domain.
            C = Y*U

        # Finally take the linear combinations of the basis for the
        # domain defined by C. Together with the kernel K, this spans
        # the inverse image of V.
        dom = self.domain()
        if not dom.is_ambient():
            C = C * dom.basis_matrix()
        L = dom.submodule(C.rows())
        return K + L

    def lift(self, x):
        r"""
        Given an element of the image, return an element of the codomain that maps onto it.

        Note that ``lift`` and ``preimage_representative`` are
        equivalent names for this method, with the latter suggesting
        that the return value is a coset representative of the domain
        modulo the kernel of the morphism.

        EXAMPLES::

            sage: X = QQ**2
            sage: V = X.span([[2, 0], [0, 8]], ZZ)
            sage: W = (QQ**1).span([[1/12]], ZZ)
            sage: f = V.hom([W([1/3]), W([1/2])], W)
            sage: l=f.lift([1/3]); l  # random
            (8, -16)
            sage: f(l)
            (1/3)
            sage: f(f.lift([1/2]))
            (1/2)
            sage: f(f.lift([1/6]))
            (1/6)
            sage: f.lift([1/12])
            Traceback (most recent call last):
            ...
            ValueError: element is not in the image
            sage: f.lift([1/24])
            Traceback (most recent call last):
            ...
            TypeError: element [1/24] is not in free module

        This works for vector spaces, too::

            sage: V = VectorSpace(GF(3), 2)
            sage: W = VectorSpace(GF(3), 3)
            sage: f = V.hom([W.1, W.1 - W.0])
            sage: f.lift(W.1)
            (1, 0)
            sage: f.lift(W.2)
            Traceback (most recent call last):
            ...
            ValueError: element is not in the image
            sage: w = W((17, -2, 0))
            sage: f(f.lift(w)) == w
            True

        This example illustrates the use of the ``preimage_representative``
        as an equivalent name for this method.  ::

            sage: V = ZZ^3
            sage: W = ZZ^2
            sage: w = vector(ZZ, [1,2])
            sage: f = V.hom([w, w, w], W)
            sage: f.preimage_representative(vector(ZZ, [10, 20]))
            (0, 0, 10)

        ::

            sage: V = QQ^2; m = matrix(2, [1, 1, 0, 1])
            sage: V.hom(m, side='right').lift(V.0 + V.1)
            (0, 1)
            sage: V.hom(m).lift(V.0 + V.1)
            (1, 0)
        """
        from .free_module_element import vector
        x = self.codomain()(x)
        if self.side() == "right":
            A = self.matrix().transpose()
        else:
            A = self.matrix()
        R = self.base_ring()
        if R.is_field():
            try:
                C = A.solve_left(x)
            except ValueError:
                raise ValueError("element is not in the image")
        else:
            # see inverse_image for similar code but with comments
            if not hasattr(A, 'hermite_form'):
                raise NotImplementedError("base ring (%s) must have hermite_form algorithm in order to compute inverse image" % R)
            H, U = A.hermite_form(transformation=True, include_zero_rows=False)
            Y = H.solve_left(vector(self.codomain().coordinates(x)))
            C = Y*U
        try:
            t = self.domain().linear_combination_of_basis(C)
        except TypeError:
            raise ValueError("element is not in the image")
        assert self(t) == x
        return t

    preimage_representative = lift

    def eigenvalues(self, extend=True):
        r"""
        Return a list with the eigenvalues of the endomorphism of vector spaces.

        INPUT:

        - ``extend`` -- boolean (default: ``True``); decides if base field
          extensions should be considered or not

        EXAMPLES:

        We compute the eigenvalues of an endomorphism of `\QQ^3`::

            sage: V = QQ^3
            sage: H = V.endomorphism_ring()([[1,-1,0], [-1,1,1], [0,3,1]])
            sage: H.eigenvalues()                                                       # needs sage.rings.number_field
            [3, 1, -1]

        Note the effect of the ``extend`` option::

            sage: V = QQ^2
            sage: H = V.endomorphism_ring()([[0,-1], [1,0]])
            sage: H.eigenvalues()                                                       # needs sage.rings.number_field
            [-1*I, 1*I]
            sage: H.eigenvalues(extend=False)                                           # needs sage.libs.pari
            []
        """
        if self.base_ring().is_field():
            if self.is_endomorphism():
                return self.matrix().eigenvalues(extend=extend)
            raise TypeError("not an endomorphism")
        else:
            raise NotImplementedError("module must be a vector space")

    def eigenvectors(self, extend=True):
        """
        Compute the subspace of eigenvectors of a given eigenvalue.

        INPUT:

        - ``extend`` -- boolean (default: ``True``); decides if base field
          extensions should be considered or not

        OUTPUT:

        A sequence of tuples. Each tuple contains an eigenvalue, a sequence
        with a basis of the corresponding subspace of eigenvectors, and the
        algebraic multiplicity of the eigenvalue.

        EXAMPLES::

            sage: # needs sage.rings.number_field
            sage: V = (QQ^4).subspace([[0,2,1,4], [1,2,5,0], [1,1,1,1]])
            sage: H = (V.Hom(V))(matrix(QQ, [[0,1,0], [-1,0,0], [0,0,3]]))
            sage: H.eigenvectors()
            [(3, [(0, 0, 1, -6/7)], 1),
             (-1*I, [(1, 1*I, 0, -0.571428571428572? + 2.428571428571429?*I)], 1),
             (1*I, [(1, -1*I, 0, -0.571428571428572? - 2.428571428571429?*I)], 1)]
            sage: H.eigenvectors(extend=False)
            [(3, [(0, 0, 1, -6/7)], 1)]
            sage: H1 = (V.Hom(V))(matrix(QQ, [[2,1,0],[0,2,0],[0,0,3]]))
            sage: H1.eigenvectors()
            [(3, [(0, 0, 1, -6/7)], 1), (2, [(0, 1, 0, 17/7)], 2)]
            sage: H1.eigenvectors(extend=False)
            [(3, [(0, 0, 1, -6/7)], 1), (2, [(0, 1, 0, 17/7)], 2)]

        ::

            sage: V = QQ^2
            sage: m = matrix(2, [1, 1, 0, 1])
            sage: V.hom(m, side='right').eigenvectors()                                 # needs sage.rings.number_field
            [(1, [(1, 0)], 2)]
            sage: V.hom(m).eigenvectors()                                               # needs sage.rings.number_field
            [(1, [(0, 1)], 2)]
        """
        if self.base_ring().is_field():
            if self.is_endomorphism():
                if self.side() == "right":
                    seigenvec = self.matrix().eigenvectors_right(extend=extend)
                else:
                    seigenvec = self.matrix().eigenvectors_left(extend=extend)
                resu = []
                for i in seigenvec:
                    V = self.domain().base_extend(i[0].parent())
                    svectors = Sequence([V(j * V.basis_matrix()) for j in i[1]], cr=True)
                    resu.append((i[0], svectors, i[2]))
                return resu
            raise TypeError("not an endomorphism")
        else:
            raise NotImplementedError("module must be a vector space")

    def eigenspaces(self, extend=True):
        """
        Compute a list of subspaces formed by eigenvectors of ``self``.

        INPUT:

        - ``extend`` -- boolean (default: ``True``); determines if field
          extensions should be considered

        OUTPUT: a list of pairs ``(eigenvalue, eigenspace)``

        EXAMPLES::

            sage: V = QQ^3
            sage: h = V.hom([[1,0,0], [0,0,1], [0,-1,0]], V)
            sage: h.eigenspaces()                                                       # needs sage.rings.number_field
            [(1,    Vector space of degree 3 and dimension 1 over Rational Field
                     Basis matrix:
                     [1 0 0]),
             (-1*I, Vector space of degree 3 and dimension 1 over Algebraic Field
                     Basis matrix:
                     [  0   1 1*I]),
             (1*I,  Vector space of degree 3 and dimension 1 over Algebraic Field
                     Basis matrix:
                     [   0    1 -1*I])]

            sage: h.eigenspaces(extend=False)                                           # needs sage.rings.number_field
            [(1,
              Vector space of degree 3 and dimension 1 over Rational Field
              Basis matrix:
              [1 0 0])]

            sage: h = V.hom([[2,1,0], [0,2,0], [0,0,-1]], V)
            sage: h.eigenspaces()                                                       # needs sage.rings.number_field
            [(-1, Vector space of degree 3 and dimension 1 over Rational Field
                   Basis matrix:
                   [0 0 1]),
             (2, Vector space of degree 3 and dimension 1 over Rational Field
                  Basis matrix:
                  [0 1 0])]

            sage: h = V.hom([[2,1,0], [0,2,0], [0,0,2]], V)
            sage: h.eigenspaces()                                                       # needs sage.rings.number_field
            [(2, Vector space of degree 3 and dimension 2 over Rational Field
                  Basis matrix:
                  [0 1 0]
                  [0 0 1])]

        ::

            sage: V = QQ^2; m = matrix(2, [1, 1, 0, 1])
            sage: V.hom(m, side='right').eigenspaces()                                  # needs sage.rings.number_field
            [(1, Vector space of degree 2 and dimension 1 over Rational Field
                  Basis matrix:
                  [1 0])]
            sage: V.hom(m).eigenspaces()                                                # needs sage.rings.number_field
            [(1, Vector space of degree 2 and dimension 1 over Rational Field
                  Basis matrix:
                  [0 1])]
        """
        ev = self.eigenvectors(extend)
        return [(vec[0], Sequence(vec[1]).universe().subspace(vec[1]))
                for vec in ev]


class BaseIsomorphism1D(Morphism):
    """
    An isomorphism between a ring and a free rank-1 module over the ring.

    EXAMPLES::

        sage: R.<x,y> = QQ[]
        sage: V, from_V, to_V = R.free_module(R)
        sage: from_V
        Isomorphism morphism:
          From: Ambient free module of rank 1 over the integral domain
                Multivariate Polynomial Ring in x, y over Rational Field
          To:   Multivariate Polynomial Ring in x, y over Rational Field
    """
    def _repr_type(self) -> str:
        r"""
        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: V, from_V, to_V = R.free_module(R)
            sage: from_V._repr_type()
            'Isomorphism'
        """
        return "Isomorphism"

    def is_injective(self) -> bool:
        r"""
        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: V, from_V, to_V = R.free_module(R)
            sage: from_V.is_injective()
            True
        """
        return True

    def is_surjective(self) -> bool:
        r"""
        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: V, from_V, to_V = R.free_module(R)
            sage: from_V.is_surjective()
            True
        """
        return True

    def _richcmp_(self, other, op):
        r"""
        EXAMPLES::

            sage: R.<x,y> = QQ[]
            sage: V, fr, to = R.free_module(R)
            sage: fr == loads(dumps(fr))
            True
        """
        if isinstance(other, BaseIsomorphism1D):
            return richcmp(self._basis, other._basis, op)
        return rich_to_bool(op, 1)


class BaseIsomorphism1D_to_FM(BaseIsomorphism1D):
    """
    An isomorphism from a ring to its 1-dimensional free module.

    INPUT:

    - ``parent`` -- the homset
    - ``basis`` -- (default: 1) an invertible element of the ring

    EXAMPLES::

        sage: R = Zmod(8)
        sage: V, from_V, to_V = R.free_module(R)
        sage: v = to_V(2); v
        (2)
        sage: from_V(v)
        2
        sage: W, from_W, to_W = R.free_module(R, basis=3)
        sage: W is V
        True
        sage: w = to_W(2); w
        (6)
        sage: from_W(w)
        2

    The basis vector has to be a unit so that the map is an isomorphism::

        sage: W, from_W, to_W = R.free_module(R, basis=4)
        Traceback (most recent call last):
        ...
        ValueError: basis element must be a unit
    """
    def __init__(self, parent, basis=None):
        """
        TESTS::

            sage: R = Zmod(8)
            sage: W, from_W, to_W = R.free_module(R, basis=3)
            sage: TestSuite(to_W).run()
        """
        Morphism.__init__(self, parent)
        self._basis = basis

    def _call_(self, x):
        """
        TESTS::

            sage: R = Zmod(8)
            sage: W, from_W, to_W = R.free_module(R, basis=3)
            sage: to_W(6) # indirect doctest
            (2)
        """
        if self._basis is not None:
            x *= self._basis
        return self.codomain()([x])


class BaseIsomorphism1D_from_FM(BaseIsomorphism1D):
    """
    An isomorphism to a ring from its 1-dimensional free module.

    INPUT:

    - ``parent`` -- the homset
    - ``basis`` -- (default: 1) an invertible element of the ring

    EXAMPLES::

        sage: R.<x> = QQ[[]]
        sage: V, from_V, to_V = R.free_module(R)
        sage: v = to_V(1+x); v
        (1 + x)
        sage: from_V(v)
        1 + x
        sage: W, from_W, to_W = R.free_module(R, basis=(1-x))
        sage: W is V
        True
        sage: w = to_W(1+x); w
        (1 - x^2)
        sage: from_W(w)
        1 + x + O(x^20)

    The basis vector has to be a unit so that the map is an isomorphism::

        sage: W, from_W, to_W = R.free_module(R, basis=x)
        Traceback (most recent call last):
        ...
        ValueError: basis element must be a unit
    """
    def __init__(self, parent, basis=None):
        """
        TESTS::

            sage: R.<x> = QQ[[]]
            sage: W, from_W, to_W = R.free_module(R, basis=(1-x))
            sage: TestSuite(from_W).run(skip='_test_nonzero_equal')
        """
        Morphism.__init__(self, parent)
        self._basis = basis

    def _call_(self, x):
        """
        TESTS::

            sage: R.<x> = QQ[[]]
            sage: W, from_W, to_W = R.free_module(R, basis=(1-x))
            sage: w = to_W(1+x); w
            (1 - x^2)
            sage: from_W(w)
            1 + x + O(x^20)
        """
        if self._basis is None:
            return x[0]
        return self.codomain()(x[0] / self._basis)



"""
Cross-platform compatibility routines and wrappers.
"""

# *****************************************************************************
#       Copyright (C) 2017 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

import os
import sys

from sage.env import SAGE_LOCAL


#################################################################
# Replacements (as needed) for Python stdlib functions to provide
# better platform compatibility
#################################################################
if sys.platform == 'darwin':
    # On OSX non-standard library paths are not automatically found by the
    # find_library implementation without setting DYLD_LIBRARY_PATH; see
    # https://github.com/sagemath/sage/issues/21399#comment:25
    from ctypes.util import find_library as _orig_find_library

    def _find_library(name):
        libdirs = []
        if SAGE_LOCAL:
            libdirs.append(os.path.join(SAGE_LOCAL, 'lib'))
        orig_dyld_library_path = os.environ.get('DYLD_LIBRARY_PATH')
        try:
            if libdirs:
                colon_sep_path = ':'.join(libdirs)
                if orig_dyld_library_path:
                    colon_sep_path += ":" + orig_dyld_library_path
                os.environ['DYLD_LIBRARY_PATH'] = colon_sep_path

            return _orig_find_library(name)
        finally:
            if libdirs:
                # Set os.environ back to what it was
                if orig_dyld_library_path is not None:
                    os.environ['DYLD_LIBRARY_PATH'] = orig_dyld_library_path
                else:
                    os.environ.pop('DYLD_LIBRARY_PATH', None)
else:
    # On other Unix-like platforms, at least where gcc is available,
    # ctypes.util.find_library works, because it takes into account where gcc
    # finds the library as if were being linked with
    from ctypes.util import find_library as _find_library


def find_library(name):
    """
    Return the shared library filename for a given library.

    The library name is given without any prefixes or suffixes--(e.g.
    just "Singular", not "libSingular", as shared library naming is
    platform-specific.

    This does ''not'' currently return the absolute path of the file on most
    platforms; see https://bugs.python.org/issue21042

    EXAMPLES::

        sage: from sage.misc.compat import find_library
        sage: find_library('giac')                                                      # needs sage.libs.giac
        '...giac...'
    """

    result = _find_library(name)
    if result:
        return result

    # if no result found, check DYLD_LIBRARY_PATH
    LDPATH_STR = os.environ.get('DYLD_LIBRARY_PATH')
    lib_dirs = (LDPATH_STR.split(':') if LDPATH_STR else []) + [os.path.join(SAGE_LOCAL, 'lib')]
    for libdir in lib_dirs:
        for libext in ['so', 'a']:
            implib = os.path.join(libdir,
                                  'lib{0}.{1}'.format(name, libext))
            if os.path.exists(implib):
                return implib



r"""
Conjugacy Classes Of The Symmetric Group

AUTHORS:

- Vincent Delecroix, Travis Scrimshaw (2014-11-23)
"""

from sage.groups.conjugacy_classes import ConjugacyClass, ConjugacyClassGAP
from sage.groups.perm_gps.permgroup_element import PermutationGroupElement
from sage.combinat.partition import Partitions_n, _Partitions
from sage.combinat.set_partition import SetPartitions
from sage.combinat.permutation import Permutation, from_cycles
from sage.sets.set import Set
import itertools


class SymmetricGroupConjugacyClassMixin:
    r"""
    Mixin class which contains methods for conjugacy classes of
    the symmetric group.
    """
    def __init__(self, domain, part):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: g = G([(1,2), (3,4,5)])
            sage: C = G.conjugacy_class(Partition([3,2]))                               # needs sage.combinat
            sage: type(C._part)                                                         # needs sage.combinat
            <class 'sage.combinat.partition.Partitions_n_with_category.element_class'>
        """
        P = Partitions_n(len(domain))
        self._part = P(part)
        self._domain = domain
        self._set = None

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: G.conjugacy_class(Partition([4]))                                     # needs sage.combinat
            Conjugacy class of cycle type [4] in
             Symmetric group of order 4! as a permutation group
        """
        return "Conjugacy class of cycle type %s in %s" % (self._part, self._parent)

    def __eq__(self, other):
        r"""
        Comparison of conjugacy classes is done by comparing the
        defining cycle types.

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: g = G([(1,2), (3,4,5)])
            sage: C = G.conjugacy_class(Partition([3,2]))                               # needs sage.combinat
            sage: Cp = G.conjugacy_class(g)                                             # needs sage.combinat
            sage: C == Cp                                                               # needs sage.combinat
            True
        """
        if not isinstance(other, SymmetricGroupConjugacyClassMixin):
            return False
        return self._part == other._part

    def __ne__(self, other):
        """
        Test for unequality.

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: g = G([(1,3), (2,4,5)])
            sage: C = G.conjugacy_class(Partition([3,2]))                               # needs sage.combinat
            sage: Cp = G.conjugacy_class(g)                                             # needs sage.combinat
            sage: C != Cp                                                               # needs sage.combinat
            False
        """
        return not (self == other)

    def partition(self):
        """
        Return the partition of ``self``.

        EXAMPLES::

            sage: G = SymmetricGroup(5)
            sage: g = G([(1,2), (3,4,5)])
            sage: C = G.conjugacy_class(g)                                              # needs sage.combinat
        """
        return self._part


class SymmetricGroupConjugacyClass(SymmetricGroupConjugacyClassMixin, ConjugacyClassGAP):
    """
    A conjugacy class of the symmetric group.

    INPUT:

    - ``group`` -- the symmetric group
    - ``part`` -- a partition or an element of ``group``
    """
    def __init__(self, group, part):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: G = SymmetricGroup(5)
            sage: g = G([(1,2), (3,4,5)])
            sage: C = G.conjugacy_class(g)
            sage: TestSuite(C).run()
            sage: C = G.conjugacy_class(Partition([3,2]))
            sage: TestSuite(C).run()
        """
        if isinstance(part, PermutationGroupElement) and part.parent() is group:
            elt = part
            part = sorted([len(x) for x in part.cycle_tuples(True)], reverse=True)
        else:
            elt = default_representative(part, group)
        SymmetricGroupConjugacyClassMixin.__init__(self, group.domain(), part)
        ConjugacyClassGAP.__init__(self, group, elt)

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: G = SymmetricGroup(4)
            sage: C = G.conjugacy_class(Partition([3,1]))                               # needs sage.combinat
            sage: for x in C: x                                                         # needs sage.combinat
            (2,3,4)
            (2,4,3)
            (1,2,3)
            (1,3,2)
            (1,2,4)
            (1,4,2)
            (1,3,4)
            (1,4,3)
        """
        if self._set:
            for x in self._set:
                yield x
            return

        for x in conjugacy_class_iterator(self._part, self._parent.domain()):
            yield self._parent.element_class(x, self._parent, check=False)

    def set(self):
        r"""
        The set of all elements in the conjugacy class ``self``.

        EXAMPLES::

            sage: G = SymmetricGroup(3)
            sage: g = G((1,2))
            sage: C = G.conjugacy_class(g)                                              # needs sage.combinat
            sage: S = [(2,3), (1,2), (1,3)]
            sage: C.set() == Set(G(x) for x in S)                                       # needs sage.combinat
            True
        """
        if not self._set:
            self._set = Set(self._parent.element_class(x, self._parent, check=False)
                            for x in conjugacy_class_iterator(self._part, self._domain))
        return self._set


class PermutationsConjugacyClass(SymmetricGroupConjugacyClassMixin, ConjugacyClass):
    """
    A conjugacy class of the permutations of `n`.

    INPUT:

    - ``P`` -- the permutations of `n`
    - ``part`` -- a partition or an element of ``P``
    """
    def __init__(self, P, part):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: # needs sage.combinat
            sage: G = Permutations(5)
            sage: g = G([2, 1, 4, 5, 3])
            sage: C = G.conjugacy_class(g)
            sage: TestSuite(C).run()
            sage: C = G.conjugacy_class(Partition([3,2]))
            sage: TestSuite(C).run()
        """
        if isinstance(part, Permutation) and part.parent() is P:
            elt = part
            part = elt.cycle_type()
        else:
            elt = P.element_in_conjugacy_classes(part)
        SymmetricGroupConjugacyClassMixin.__init__(self, range(1, P.n + 1), part)
        ConjugacyClass.__init__(self, P, elt)

    def __iter__(self):
        """
        Iterate over ``self``.

        EXAMPLES::

            sage: G = Permutations(4)
            sage: C = G.conjugacy_class(Partition([3,1]))                               # needs sage.combinat
            sage: for x in C: x                                                         # needs sage.combinat
            [1, 3, 4, 2]
            [1, 4, 2, 3]
            [2, 3, 1, 4]
            [3, 1, 2, 4]
            [2, 4, 3, 1]
            [4, 1, 3, 2]
            [3, 2, 4, 1]
            [4, 2, 1, 3]
        """
        if self._set:
            for x in self._set:
                yield x
            return

        for x in conjugacy_class_iterator(self._part, self._domain):
            yield from_cycles(self._parent.n, x, self._parent)

    def set(self):
        r"""
        The set of all elements in the conjugacy class ``self``.

        EXAMPLES::

            sage: G = Permutations(3)
            sage: g = G([2, 1, 3])
            sage: C = G.conjugacy_class(g)                                              # needs sage.combinat
            sage: S = [[1, 3, 2], [2, 1, 3], [3, 2, 1]]
            sage: C.set() == Set(G(x) for x in S)                                       # needs sage.combinat
            True
        """
        if not self._set:
            self._set = Set(from_cycles(self._parent.n, x, self._parent)
                            for x in conjugacy_class_iterator(self._part, self._domain))
        return self._set


#####################################################################
# Helper functions

def default_representative(part, G):
    r"""
    Construct the default representative for the conjugacy class of
    cycle type ``part`` of a symmetric group ``G``.

    Let `\lambda` be a partition of `n`. We pick a representative by

    .. MATH::

        (1, 2, \ldots, \lambda_1)
        (\lambda_1 + 1, \ldots, \lambda_1 + \lambda_2)
        (\lambda_1 + \lambda_2 + \cdots + \lambda_{\ell-1}, \ldots, n),

    where `\ell` is the length (or number of parts) of `\lambda`.

    INPUT:

    - ``part`` -- partition

    - ``G`` -- a symmetric group

    EXAMPLES::

        sage: from sage.groups.perm_gps.symgp_conjugacy_class import default_representative         # needs sage.combinat
        sage: S = SymmetricGroup(4)
        sage: for p in Partitions(4):                                                   # needs sage.combinat
        ....:     print(default_representative(p, S))
        (1,2,3,4)
        (1,2,3)
        (1,2)(3,4)
        (1,2)
        ()
    """
    D = G.domain()
    total = 0
    cycles = []
    for p in part:
        cycles.append(tuple(D[total:total + p]))
        total += p
    return G.element_class(cycles, G, check=False)


def conjugacy_class_iterator(part, S=None):
    r"""
    Return an iterator over the conjugacy class associated to
    the partition ``part``.

    The elements are given as a list of tuples, each tuple being a cycle.

    INPUT:

    - ``part`` -- partition

    - ``S`` -- (default: `\{ 1, 2, \ldots, n \}`, where `n`
      is the size of ``part``) a set

    OUTPUT: an iterator over the conjugacy class consisting of all
    permutations of the set ``S`` whose cycle type is ``part``

    EXAMPLES::

        sage: from sage.groups.perm_gps.symgp_conjugacy_class import conjugacy_class_iterator       # needs sage.combinat
        sage: for p in conjugacy_class_iterator([2,2]): print(p)                        # needs sage.combinat
        [(1, 2), (3, 4)]
        [(1, 4), (2, 3)]
        [(1, 3), (2, 4)]

    In order to get permutations, one just has to wrap::

        sage: S = SymmetricGroup(5)
        sage: for p in conjugacy_class_iterator([3,2]): print(S(p))                     # needs sage.combinat
        (1,3)(2,4,5)
        (1,3)(2,5,4)
        (1,2)(3,4,5)
        (1,2)(3,5,4)
        ...
        (1,4)(2,3,5)
        (1,4)(2,5,3)

    Check that the number of elements is the number of elements in
    the conjugacy class::

        sage: s = lambda p: sum(1 for _ in conjugacy_class_iterator(p))
        sage: all(s(p) == p.conjugacy_class_size() for p in Partitions(5))              # needs sage.combinat
        True

    It is also possible to specify any underlying set::

        sage: it = conjugacy_class_iterator([2,2,2], 'abcdef')                          # needs sage.combinat
        sage: sorted(flatten(next(it)))                                                 # needs sage.combinat
        ['a', 'b', 'c', 'd', 'e', 'f']
        sage: all(len(x) == 2 for x in next(it))                                        # needs sage.combinat
        True
    """
    n = sum(part)
    if part not in _Partitions:
        raise ValueError("invalid partition")
    if S is None:
        S = range(1, n + 1)
    else:
        S = list(S)
        if n != len(S):
            raise ValueError("the sum of the partition %s does not match the size of %s" % (part, S))

    m = len(part)
    for s in SetPartitions(S, part):
        its = [iter(t) for t in s]
        firsts = [next(t) for t in its]
        rests = [list(t) for t in its]
        iterator = tuple(itertools.permutations(r) for r in rests)
        for r in itertools.product(*iterator):
            yield [(firsts[i],) + r[i] for i in range(m)]
